'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var EPSILON = Number.EPSILON || 2.2204460492503130808472633361816E-16;
function createAkimaSplineInterpolator(xVals, yVals) {
    var segmentCoeffs = computeAkimaPolyCoefficients(xVals, yVals);
    var xValsCopy = Float64Array.from(xVals);
    return function (x) { return evaluatePolySegment(xValsCopy, segmentCoeffs, x); };
}
function computeAkimaPolyCoefficients(xVals, yVals) {
    if (xVals.length != yVals.length) {
        throw new Error("Dimension mismatch for xVals and yVals.");
    }
    if (xVals.length < 5) {
        throw new Error("Number of points is too small.");
    }
    MathArrays_checkOrder(xVals);
    var n = xVals.length - 1;
    var differences = new Float64Array(n);
    var weights = new Float64Array(n);
    for (var i = 0; i < n; i++) {
        differences[i] = (yVals[i + 1] - yVals[i]) / (xVals[i + 1] - xVals[i]);
    }
    for (var i = 1; i < n; i++) {
        weights[i] = Math.abs(differences[i] - differences[i - 1]);
    }
    var firstDerivatives = new Float64Array(n + 1);
    for (var i = 2; i < n - 1; i++) {
        var wP = weights[i + 1];
        var wM = weights[i - 1];
        if (Math.abs(wP) < EPSILON && Math.abs(wM) < EPSILON) {
            var xv = xVals[i];
            var xvP = xVals[i + 1];
            var xvM = xVals[i - 1];
            firstDerivatives[i] = (((xvP - xv) * differences[i - 1]) + ((xv - xvM) * differences[i])) / (xvP - xvM);
        }
        else {
            firstDerivatives[i] = ((wP * differences[i - 1]) + (wM * differences[i])) / (wP + wM);
        }
    }
    firstDerivatives[0] = differentiateThreePoint(xVals, yVals, 0, 0, 1, 2);
    firstDerivatives[1] = differentiateThreePoint(xVals, yVals, 1, 0, 1, 2);
    firstDerivatives[n - 1] = differentiateThreePoint(xVals, yVals, n - 1, n - 2, n - 1, n);
    firstDerivatives[n] = differentiateThreePoint(xVals, yVals, n, n - 2, n - 1, n);
    return computeHermitePolyCoefficients(xVals, yVals, firstDerivatives);
}
function differentiateThreePoint(xVals, yVals, indexOfDifferentiation, indexOfFirstSample, indexOfSecondsample, indexOfThirdSample) {
    var x0 = yVals[indexOfFirstSample];
    var x1 = yVals[indexOfSecondsample];
    var x2 = yVals[indexOfThirdSample];
    var t = xVals[indexOfDifferentiation] - xVals[indexOfFirstSample];
    var t1 = xVals[indexOfSecondsample] - xVals[indexOfFirstSample];
    var t2 = xVals[indexOfThirdSample] - xVals[indexOfFirstSample];
    var a = (x2 - x0 - (t2 / t1 * (x1 - x0))) / (t2 * t2 - t1 * t2);
    var b = (x1 - x0 - a * t1 * t1) / t1;
    return (2 * a * t) + b;
}
function computeHermitePolyCoefficients(xVals, yVals, firstDerivatives) {
    if (xVals.length != yVals.length || xVals.length != firstDerivatives.length) {
        throw new Error("Dimension mismatch");
    }
    if (xVals.length < 2) {
        throw new Error("Not enough points.");
    }
    var n = xVals.length - 1;
    var segmentCoeffs = new Array(n);
    for (var i = 0; i < n; i++) {
        var w = xVals[i + 1] - xVals[i];
        var w2 = w * w;
        var yv = yVals[i];
        var yvP = yVals[i + 1];
        var fd = firstDerivatives[i];
        var fdP = firstDerivatives[i + 1];
        var coeffs = new Float64Array(4);
        coeffs[0] = yv;
        coeffs[1] = firstDerivatives[i];
        coeffs[2] = (3 * (yvP - yv) / w - 2 * fd - fdP) / w;
        coeffs[3] = (2 * (yv - yvP) / w + fd + fdP) / w2;
        segmentCoeffs[i] = trimPoly(coeffs);
    }
    return segmentCoeffs;
}
function createCubicSplineInterpolator(xVals, yVals) {
    var segmentCoeffs = computeCubicPolyCoefficients(xVals, yVals);
    var xValsCopy = Float64Array.from(xVals);
    return function (x) { return evaluatePolySegment(xValsCopy, segmentCoeffs, x); };
}
function computeCubicPolyCoefficients(xVals, yVals) {
    if (xVals.length != yVals.length) {
        throw new Error("Dimension mismatch.");
    }
    if (xVals.length < 3) {
        throw new Error("Number of points is too small.");
    }
    MathArrays_checkOrder(xVals);
    var n = xVals.length - 1;
    var h = new Float64Array(n);
    for (var i = 0; i < n; i++) {
        h[i] = xVals[i + 1] - xVals[i];
    }
    var mu = new Float64Array(n);
    var z = new Float64Array(n + 1);
    mu[0] = 0;
    z[0] = 0;
    for (var i = 1; i < n; i++) {
        var g = 2 * (xVals[i + 1] - xVals[i - 1]) - h[i - 1] * mu[i - 1];
        mu[i] = h[i] / g;
        z[i] = (3 * (yVals[i + 1] * h[i - 1] - yVals[i] * (xVals[i + 1] - xVals[i - 1]) + yVals[i - 1] * h[i]) /
            (h[i - 1] * h[i]) - h[i - 1] * z[i - 1]) / g;
    }
    var b = new Float64Array(n);
    var c = new Float64Array(n + 1);
    var d = new Float64Array(n);
    z[n] = 0;
    c[n] = 0;
    for (var i = n - 1; i >= 0; i--) {
        var dx = h[i];
        var dy = yVals[i + 1] - yVals[i];
        c[i] = z[i] - mu[i] * c[i + 1];
        b[i] = dy / dx - dx * (c[i + 1] + 2 * c[i]) / 3;
        d[i] = (c[i + 1] - c[i]) / (3 * dx);
    }
    var segmentCoeffs = new Array(n);
    for (var i = 0; i < n; i++) {
        var coeffs = new Float64Array(4);
        coeffs[0] = yVals[i];
        coeffs[1] = b[i];
        coeffs[2] = c[i];
        coeffs[3] = d[i];
        segmentCoeffs[i] = trimPoly(coeffs);
    }
    return segmentCoeffs;
}
function createLinearInterpolator(xVals, yVals) {
    var segmentCoeffs = computeLinearPolyCoefficients(xVals, yVals);
    var xValsCopy = Float64Array.from(xVals);
    return function (x) { return evaluatePolySegment(xValsCopy, segmentCoeffs, x); };
}
function computeLinearPolyCoefficients(xVals, yVals) {
    if (xVals.length != yVals.length) {
        throw new Error("Dimension mismatch.");
    }
    if (xVals.length < 2) {
        throw new Error("Number of points is too small.");
    }
    MathArrays_checkOrder(xVals);
    var n = xVals.length - 1;
    var segmentCoeffs = new Array(n);
    for (var i = 0; i < n; i++) {
        var dx = xVals[i + 1] - xVals[i];
        var dy = yVals[i + 1] - yVals[i];
        var m = dy / dx;
        var c = new Float64Array(2);
        c[0] = yVals[i];
        c[1] = m;
        segmentCoeffs[i] = trimPoly(c);
    }
    return segmentCoeffs;
}
function createNearestNeighborInterpolator(xVals, yVals) {
    var xVals2 = Float64Array.from(xVals);
    var yVals2 = Float64Array.from(yVals);
    var n = xVals2.length;
    if (n != yVals2.length) {
        throw new Error("Dimension mismatch for xVals and yVals.");
    }
    if (n == 0) {
        return function (_x) {
            return NaN;
        };
    }
    if (n == 1) {
        return function (_x) {
            return yVals2[0];
        };
    }
    MathArrays_checkOrder(xVals2);
    return function (x) {
        var i = Arrays_binarySearch(xVals2, x);
        if (i >= 0) {
            return yVals2[i];
        }
        i = -i - 1;
        if (i == 0) {
            return yVals2[0];
        }
        if (i >= n) {
            return yVals2[n - 1];
        }
        var d = x - xVals2[i - 1];
        var w = xVals2[i] - xVals2[i - 1];
        return (d + d < w) ? yVals2[i - 1] : yVals2[i];
    };
}
function evaluatePolySegment(xVals, segmentCoeffs, x) {
    var i = Arrays_binarySearch(xVals, x);
    if (i < 0) {
        i = -i - 2;
    }
    i = Math.max(0, Math.min(i, segmentCoeffs.length - 1));
    return evaluatePoly(segmentCoeffs[i], x - xVals[i]);
}
function evaluatePoly(c, x) {
    var n = c.length;
    if (n == 0) {
        return 0;
    }
    var v = c[n - 1];
    for (var i = n - 2; i >= 0; i--) {
        v = x * v + c[i];
    }
    return v;
}
function trimPoly(c) {
    var n = c.length;
    while (n > 1 && c[n - 1] == 0) {
        n--;
    }
    return (n == c.length) ? c : c.subarray(0, n);
}
function MathArrays_checkOrder(a) {
    for (var i = 1; i < a.length; i++) {
        if (a[i] <= a[i - 1]) {
            throw new Error("Non-monotonic sequence exception.");
        }
    }
}
function Arrays_binarySearch(a, key) {
    var low = 0;
    var high = a.length - 1;
    while (low <= high) {
        var mid = (low + high) >>> 1;
        var midVal = a[mid];
        if (midVal < key) {
            low = mid + 1;
        }
        else if (midVal > key) {
            high = mid - 1;
        }
        else if (midVal == key) {
            return mid;
        }
        else {
            throw new Error("Invalid number encountered in binary search.");
        }
    }
    return -(low + 1);
}

exports.computeAkimaPolyCoefficients = computeAkimaPolyCoefficients;
exports.computeCubicPolyCoefficients = computeCubicPolyCoefficients;
exports.computeLinearPolyCoefficients = computeLinearPolyCoefficients;
exports.createAkimaSplineInterpolator = createAkimaSplineInterpolator;
exports.createCubicSplineInterpolator = createCubicSplineInterpolator;
exports.createLinearInterpolator = createLinearInterpolator;
exports.createNearestNeighborInterpolator = createNearestNeighborInterpolator;
